#! /usr/bin/env python3
##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import sys
import os
import time
import argparse
import subprocess
import signal
import datetime
import xml.etree.ElementTree as ET

class colors:
    FAILURE = '\033[1;31m' # red
    SUCCESS = '\033[1;32m' # green
    INFO    = '\033[1;33m' # yellow
    PREFIX  = '\033[1;36m' # cyan
    OTHER   = '\033[1;35m' # purple
    END     = '\033[0m'    # reset

opts = {"verbose": 0}

# junit variables
num_tests = 0
num_failures = 0
testnames = []
testtimes = []
testretvals = []
testoutputs = []


def init_colors():
    if not sys.stdout.isatty():
        colors.FAILURE = ''
        colors.SUCCESS = ''
        colors.INFO = ''
        colors.PREFIX = ''
        colors.OTHER = ''
        colors.END = ''


def printout_make(line, ret, elapsed_time, output):
    global num_tests
    global num_failures

    sys.stdout.write(colors.PREFIX + ">>>> " + colors.END)
    sys.stdout.write("return status: ")
    if (ret == 0):
        sys.stdout.write(colors.SUCCESS + "SUCCESS\n" + colors.END)
    else:
        sys.stdout.write(colors.FAILURE + "FAILURE\n" + colors.END)
    sys.stdout.write(colors.PREFIX + ">>>> " + colors.END)
    sys.stdout.write("elapsed time: %f sec\n" % elapsed_time)
    if (ret != 0):
        num_tests = num_tests + 1
        num_failures = num_failures + 1
        if (output != ""):
            execname = line.split(' ', 1)[0].rstrip()
            print(colors.FAILURE + "\n==== \"make %s\" failed ====" % execname + colors.END)
            print(output)
            print(colors.FAILURE + "==== make output complete ====\n" + colors.END)
        testnames.append(line)
        testtimes.append(elapsed_time)
        testretvals.append(ret)
        testoutputs.append(output)


def printout_exec(line, ret, elapsed_time, output):
    global num_tests
    global num_failures

    num_tests = num_tests + 1
    sys.stdout.write(colors.PREFIX + ">>>> " + colors.END)
    sys.stdout.write("return status: ")
    if (ret == 0):
        sys.stdout.write(colors.SUCCESS + "SUCCESS\n" + colors.END)
    else:
        sys.stdout.write(colors.FAILURE + "FAILURE\n" + colors.END)
        num_failures = num_failures + 1
    sys.stdout.write(colors.PREFIX + ">>>> " + colors.END)
    sys.stdout.write("elapsed time: %f sec\n" % elapsed_time)
    if (ret != 0):
        if (output != ""):
            print(colors.FAILURE + "==== execution failed with the following output ====" + colors.END)
            print(output)
            print(colors.FAILURE + "==== execution output complete ====\n" + colors.END)

    testnames.append(line)
    testtimes.append(elapsed_time)
    testretvals.append(ret)
    testoutputs.append(output)


def getlines(fh):
    alllines = fh.readlines()
    reallines = []
    for line in alllines:
        # skip comments
        if line.startswith("#"):
            continue
        # skip empty lines
        if not line.strip():
            continue
        reallines.append(line)
    return reallines


def create_summary(summary_file):
    global num_tests
    global num_failures

    # open the summary file and write to it
    try:
        fh = open(summary_file, "w")
    except:
        sys.stderr.write(colors.FAILURE + ">>>> ERROR: " + colors.END)
        sys.stderr.write("could not open summary file %s\n" % summary_file)
        sys.exit()
    fh.write("<testsuites>\n")
    fh.write("  <testsuite failures=\"%d\"\n" % num_failures)
    fh.write("             errors=\"0\"\n")
    fh.write("             skipped=\"0\"\n")
    fh.write("             tests=\"%d\"\n" % num_tests)
    fh.write("             date=\"%s\"\n" % datetime.datetime.now())
    fh.write("             name=\"summary_junit_xml\">\n")

    for x in range(len(testnames)):
        fh.write("    <testcase name=\"%s\" time=\"%f\">\n" % (testnames[x].strip(), testtimes[x]))
        if (testretvals[x] != 0):
            fh.write("      <failure><![CDATA[\n")
            if (testoutputs[x]):
                fh.write(testoutputs[x] + "\n")
            else:
                fh.write("test failed\n")
            fh.write("      ]]></failure>\n")
        fh.write("    </testcase>\n")

    fh.write("  </testsuite>\n")
    fh.write("</testsuites>\n")
    fh.close()


def wait_with_signal(p):
    try:
        ret = p.wait()
    except:
        p.kill()
        p.wait()
        sys.exit()
    return ret


def run_testlist(testlist):
    try:
        fh = open(testlist, "r")
    except:
        sys.stderr.write(colors.FAILURE + ">>>> ERROR: " + colors.END)
        sys.stderr.write("could not open testlist %s\n" % testlist)
        sys.exit()

    print(colors.INFO + "\n==== executing testlist %s ====" % testlist + colors.END)

    lines = getlines(fh)

    firstline = 1
    for line in lines:
        if (firstline):
            firstline = 0
        else:
            print("")


        ############################################################################
        # if the first argument is a directory, step into the
        # directory and reexecute make
        ############################################################################
        dirname = line.split(' ', 1)[0].rstrip()
        if (os.path.isdir(dirname)):
            sys.stdout.write(colors.PREFIX + ">>>> " + colors.END)
            sys.stdout.write(colors.OTHER + "stepping into directory %s\n" % dirname + colors.END)
            olddir = os.getcwd()
            os.chdir(dirname)
            chdirargs = "make -s testing".split(' ')
            chdirargs = map(lambda s: s.strip(), chdirargs)
            p = subprocess.Popen(chdirargs)
            wait_with_signal(p)
            os.chdir(olddir)
            continue


        # command line to process
        sys.stdout.write(line)


        ############################################################################
        # make executable
        ############################################################################
        execname = line.split(' ', 1)[0].rstrip()
        if opts['verbose']:
            print("make " + execname)
        start = time.time()
        p = subprocess.Popen(['make', execname], stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
        ret = wait_with_signal(p)
        out = p.communicate()
        end = time.time()
        if opts['verbose']:
            print(out[0].decode().strip())
        printout_make(line, ret, end - start, out[0].decode().strip())
        if (ret != 0):
            continue  # skip over to the next line


        ############################################################################
        # run the executable
        ############################################################################
        fullcmd = "./" + line
        cmdargs = fullcmd.split(' ')
        cmdargs = map(lambda s: s.strip(), cmdargs)
        start = time.time()
        p = subprocess.Popen(cmdargs, stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
        ret = wait_with_signal(p)
        out = p.communicate()
        end = time.time()
        printout_exec(line, ret, end - start, out[0].decode().strip())

    fh.close()
    print(colors.INFO + "==== done executing testlist %s ====" % testlist + colors.END)


if __name__ == '__main__':
    init_colors()

    parser = argparse.ArgumentParser()
    parser.add_argument('testlists', help='testlist files to execute', nargs='+')
    parser.add_argument('--summary', help='file to write the summary to', required=True)
    args = parser.parse_args()

    if os.environ.get('V'):
        opts["verbose"] = 1

    for testlist in args.testlists:
        run_testlist(os.path.abspath(testlist))
    create_summary(os.path.abspath(args.summary))





#! /usr/bin/env python3
##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import sys

########################################################################################
##### printing utilities
########################################################################################
def display(OUTFILE, *argv):
    for arg in argv:
        if (arg.find('}') != -1):
            display.indent -= 1
    for x in range(display.indent):
        OUTFILE.write("    ")
    for arg in argv:
        OUTFILE.write(arg)
        if (arg.find('{') != -1):
            display.indent += 1
display.indent = 0


########################################################################################
##### add the copyright header to the top of the file
########################################################################################
def copyright_c(outfile):
    OUTFILE = open(outfile, "w")
    OUTFILE.write("/*\n")
    OUTFILE.write(" * Copyright (C) by Argonne National Laboratory\n")
    OUTFILE.write(" *     See COPYRIGHT in top-level directory\n")
    OUTFILE.write(" *\n")
    OUTFILE.write(" * DO NOT EDIT: AUTOMATICALLY GENERATED FILE !!\n")
    OUTFILE.write(" */\n")
    OUTFILE.write("\n")
    OUTFILE.close()

def copyright_makefile(outfile):
    OUTFILE = open(outfile, "w")
    OUTFILE.write("##\n")
    OUTFILE.write("## Copyright (C) by Argonne National Laboratory\n")
    OUTFILE.write("##     See COPYRIGHT in top-level directory\n")
    OUTFILE.write("##\n")
    OUTFILE.write("## DO NOT EDIT: AUTOMATICALLY GENERATED FILE !!\n")
    OUTFILE.write("##\n")
    OUTFILE.write("\n")
    OUTFILE.close()


########################################################################################
##### generate an array of datatype arrays
########################################################################################
def generate_darrays(derived_types, darraylist, maxlevels):
    darraylist.append([])
    for level in range(maxlevels, 0, -1):
        index = [ ]
        for x in range(level):
            index.append(0)

        while True:
            darray = [ ]
            for x in range(level):
                darray.append(derived_types[index[x]])
            darraylist.append(darray)

            index[-1] = index[-1] + 1
            for x in range(level - 1, 0, -1):
                if (index[x] == len(derived_types)):
                    index[x] = 0
                    index[x-1] = index[x-1] + 1
            if (index[0] == len(derived_types)):
                break



#! /usr/bin/env python3
##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import sys
import os
import argparse
import subprocess
from collections import deque

sys.path.append('maint/')
import yutils

sys.path.append('src/backend/')
import gencomm


########################################################################################
##### Global variables
########################################################################################

num_paren_open = 0
blklens = [ "generic" ]
builtin_types = [ "char", "int", "short", "long", "int8_t", "int16_t", \
                  "int32_t", "int64_t", "float", "double" ]
builtin_maps = {
    "YAKSA_TYPE__UNSIGNED_CHAR": "char",
    "YAKSA_TYPE__UNSIGNED": "int",
    "YAKSA_TYPE__UNSIGNED_SHORT": "short",
    "YAKSA_TYPE__UNSIGNED_LONG": "long",
    "YAKSA_TYPE__LONG_DOUBLE": "double",
    "YAKSA_TYPE__UNSIGNED_LONG_LONG": "long",
    "YAKSA_TYPE__UINT8_T": "int8_t",
    "YAKSA_TYPE__UINT16_T": "int16_t",
    "YAKSA_TYPE__UINT32_T": "int32_t",
    "YAKSA_TYPE__UINT64_T": "int64_t",
    "YAKSA_TYPE__C_COMPLEX": "float",
    "YAKSA_TYPE__C_DOUBLE_COMPLEX": "double",
    "YAKSA_TYPE__C_LONG_DOUBLE_COMPLEX": "double",
    "YAKSA_TYPE__BYTE": "int8_t"
}


########################################################################################
##### Type-specific functions
########################################################################################

## hvector routines
def hvector(suffix, dtp, b, last):
    global s
    global idx
    global need_extent
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.hvector.stride;\n" % (suffix, dtp))
    if (need_extent == True):
        yutils.display(OUTFILE, "uintptr_t extent%d = %s->extent;\n" % (suffix, dtp))
    if (last != 1):
        s += " + x%d * stride%d + x%d * extent%d" % (idx, suffix, idx + 1, suffix + 1)
        need_extent = True
    else:
        s += " + x%d * stride%d + x%d * sizeof(%s)" % (idx, suffix, idx + 1, b)
        need_extent = False
    idx = idx + 2

## blkhindx routines
def blkhindx(suffix, dtp, b, last):
    global s
    global idx
    global need_extent
    yutils.display(OUTFILE, "intptr_t *array_of_displs%d = %s->u.blkhindx.array_of_displs;\n" % (suffix, dtp))
    if (need_extent == True):
        yutils.display(OUTFILE, "uintptr_t extent%d = %s->extent;\n" % (suffix, dtp))
    if (last != 1):
        s += " + array_of_displs%d[x%d] + x%d * extent%d" % \
             (suffix, idx, idx + 1, suffix + 1)
        need_extent = True
    else:
        s += " + array_of_displs%d[x%d] + x%d * sizeof(%s)" % (suffix, idx, idx + 1, b)
        need_extent = False
    idx = idx + 2

## hindexed routines
def hindexed(suffix, dtp, b, last):
    global s
    global idx
    global need_extent
    yutils.display(OUTFILE, "intptr_t *array_of_displs%d = %s->u.hindexed.array_of_displs;\n" % (suffix, dtp))
    if (need_extent == True):
        yutils.display(OUTFILE, "uintptr_t extent%d = %s->extent;\n" % (suffix, dtp))
    if (last != 1):
        s += " + array_of_displs%d[x%d] + x%d * extent%d" % \
             (suffix, idx, idx + 1, suffix + 1)
        need_extent = True
    else:
        s += " + array_of_displs%d[x%d] + x%d * sizeof(%s)" % (suffix, idx, idx + 1, b)
        need_extent = False
    idx = idx + 2

## contig routines
def contig(suffix, dtp, b, last):
    global s
    global idx
    global need_extent
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.contig.child->extent;\n" % (suffix, dtp))
    if (need_extent == True):
        yutils.display(OUTFILE, "uintptr_t extent%d = %s->extent;\n" % (suffix, dtp))
    need_extent = False
    s += " + x%d * stride%d" % (idx, suffix)
    idx = idx + 1

# resized routines
def resized(suffix, dtp, b, last):
    global need_extent
    if (need_extent == True):
        yutils.display(OUTFILE, "uintptr_t extent%d = %s->extent;\n" % (suffix, dtp))
    need_extent = False


########################################################################################
##### Core kernels
########################################################################################
def generate_kernels(b, darray):
    global need_extent
    global s
    global idx

    # we don't need pup kernels for basic types
    if (len(darray) == 0):
        return

    for func in "pack","unpack":
        ##### figure out the function name to use
        funcprefix = "%s_" % func
        for d in darray:
            funcprefix = funcprefix + "%s_" % d
        funcprefix = funcprefix + b.replace(" ", "_")

        ##### generate the ZE kernel
        yutils.display(OUTFILE, "__kernel void yaksuri_zei_kernel_%s(__global const void *inbuf, __global void *outbuf, unsigned long count, __global const yaksuri_zei_md_s *__restrict__ md)\n" % funcprefix)
        yutils.display(OUTFILE, "{\n")
        yutils.display(OUTFILE, "__global const char *__restrict__ sbuf = (__global char *) inbuf;\n");
        yutils.display(OUTFILE, "__global char *__restrict__ dbuf = (__global char *) outbuf;\n")
        if func == "pack":
            yutils.display(OUTFILE, "sbuf = (__global const char *) ((__global char *)sbuf - md->true_lb);\n");
        else:
            yutils.display(OUTFILE, "dbuf = dbuf - md->true_lb;\n");
        yutils.display(OUTFILE, "uintptr_t extent = md->extent;\n")
        yutils.display(OUTFILE, "uintptr_t idx = get_global_id(0);\n")
        yutils.display(OUTFILE, "uintptr_t res = idx;\n")
        yutils.display(OUTFILE, "uintptr_t inner_elements = md->num_elements;\n")
        yutils.display(OUTFILE, "\n")

        yutils.display(OUTFILE, "if (idx >= (count * inner_elements))\n")
        yutils.display(OUTFILE, "    return;\n")
        yutils.display(OUTFILE, "\n")

        # copy loop
        idx = 0
        md = "md"
        for d in darray:
            if (d == "hvector" or d == "blkhindx" or d == "hindexed" or \
                d == "contig"):
                yutils.display(OUTFILE, "uintptr_t x%d = res / inner_elements;\n" % idx)
                idx = idx + 1
                yutils.display(OUTFILE, "res %= inner_elements;\n")
                yutils.display(OUTFILE, "inner_elements /= %s->u.%s.count;\n" % (md, d))
                yutils.display(OUTFILE, "\n")

            if (d == "hvector" or d == "blkhindx"):
                yutils.display(OUTFILE, "uintptr_t x%d = res / inner_elements;\n" % idx)
                idx = idx + 1
                yutils.display(OUTFILE, "res %= inner_elements;\n")
                yutils.display(OUTFILE, "inner_elements /= %s->u.%s.blocklength;\n" % (md, d))
            elif (d == "hindexed"):
                yutils.display(OUTFILE, "uintptr_t x%d;\n" % idx)
                yutils.display(OUTFILE, "for (int i = 0; i < %s->u.%s.count; i++) {\n" % (md, d))
                yutils.display(OUTFILE, "    uintptr_t in_elems = %s->u.%s.array_of_blocklengths[i] *\n" % (md, d))
                yutils.display(OUTFILE, "                         %s->u.%s.child->num_elements;\n" % (md, d))
                yutils.display(OUTFILE, "    if (res < in_elems) {\n")
                yutils.display(OUTFILE, "        x%d = i;\n" % idx)
                yutils.display(OUTFILE, "        res %= in_elems;\n")
                yutils.display(OUTFILE, "        inner_elements = %s->u.%s.child->num_elements;\n" % (md, d))
                yutils.display(OUTFILE, "        break;\n")
                yutils.display(OUTFILE, "    } else {\n")
                yutils.display(OUTFILE, "        res -= in_elems;\n")
                yutils.display(OUTFILE, "    }\n")
                yutils.display(OUTFILE, "}\n")
                idx = idx + 1
                yutils.display(OUTFILE, "\n")

            md = "%s->u.%s.child" % (md, d)

        yutils.display(OUTFILE, "uintptr_t x%d = res;\n" % idx)
        yutils.display(OUTFILE, "\n")

        dtp = "md"
        s = "x0 * extent"
        idx = 1
        x = 1
        need_extent = False
        for d in darray:
            if (x == len(darray)):
                last = 1
            else:
                last = 0
            getattr(sys.modules[__name__], d)(x, dtp, b, last)
            x = x + 1
            dtp = dtp + "->u.%s.child" % d

        if (func == "pack"):
            yutils.display(OUTFILE, "*((%s *) (void *) (dbuf + idx * sizeof(%s))) = *((const %s *) (const void *) (sbuf + %s));\n"
                                       % (b, b, b, s))
        else:
            yutils.display(OUTFILE, "*((%s *) (void *) (dbuf + %s)) = *((const %s *) (const void *) (sbuf + idx * sizeof(%s)));\n"
                                       % (b, s, b, b))

        yutils.display(OUTFILE, "}\n\n")


def write_headers():
    yutils.display(OUTFILE, "typedef signed char int8_t;\n")
    yutils.display(OUTFILE, "typedef signed short int int16_t;\n")
    yutils.display(OUTFILE, "typedef signed int int32_t;\n")
    yutils.display(OUTFILE, "typedef signed long int64_t;\n")
    yutils.display(OUTFILE, "typedef unsigned char uint8_t;\n")
    yutils.display(OUTFILE, "typedef unsigned short int uint16_t;\n")
    yutils.display(OUTFILE, "typedef unsigned int uint32_t;\n")
    yutils.display(OUTFILE, "typedef unsigned long uint64_t;\n")
    yutils.display(OUTFILE, "#include \"yaksuri_zei_md.h\"\n")
    yutils.display(OUTFILE, "\n")

def generate_define_kernels(b, darray):
    global k
    for func in "pack","unpack":
        ##### figure out the function name to use
        s = "yaksuri_zei_%s_" % func
        for d in darray:
            s = s + "%s_" % d
        s = s + b.replace(" ", "_")
        OUTFILE.write("#define %s  %d\n" % (s, k))
        k += 1

########################################################################################
##### main function
########################################################################################
if __name__ == '__main__':
    ##### parse user arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--pup-max-nesting', type=int, default=3, help='maximum nesting levels to generate')
    args = parser.parse_args()
    if (args.pup_max_nesting < 0):
        parser.print_help()
        print
        print("===> ERROR: pup-max-nesting must be positive")
        sys.exit(1)

    ##### generate the core pack/unpack kernels
    darraylist = [ ]
    yutils.generate_darrays(gencomm.derived_types, darraylist, args.pup_max_nesting - 2)

    for b in builtin_types:
        for d1 in gencomm.derived_types:
            ##### generate the core pack/unpack kernels (single level)
            filename = "src/backend/ze/pup/yaksuri_zei_pup_%s_%s.cl" % (d1, b.replace(" ","_"))
            yutils.copyright_c(filename)
            OUTFILE = open(filename, "a")
            write_headers()

            emptylist = [ ]
            emptylist.append(d1)
            generate_kernels(b, emptylist)
            emptylist.pop()
            OUTFILE.close()

            ##### generate the core pack/unpack kernels (more than one level)
            for d2 in gencomm.derived_types:
                filename = "src/backend/ze/pup/yaksuri_zei_pup_%s_%s_%s.cl" % (d1, d2, b.replace(" ","_"))
                yutils.copyright_c(filename)
                OUTFILE = open(filename, "a")
                write_headers()

                for darray in darraylist:
                    darray.append(d1)
                    darray.append(d2)
                    generate_kernels(b, darray)
                    darray.pop()
                    darray.pop()

                OUTFILE.close()

    ##### generate code to load modules/kernels used by init_hook
    filename = "src/backend/ze/hooks/yaksuri_zei_init_kernels.c"
    yutils.copyright_c(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "#include \"stdlib.h\"\n")
    yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
    yutils.display(OUTFILE, "#include \"yaksuri_zei.h\"\n")
    yutils.display(OUTFILE, "#include \"level_zero/ze_api.h\"\n\n")

    num_modules = 0
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            ##### generate the core pack/unpack kernels (single level)
            OUTFILE.write("extern const unsigned char yaksuri_zei_pup_%s_%s_str[];\n" % (d1, b.replace(" ", "_")))
            OUTFILE.write("extern const size_t yaksuri_zei_pup_%s_%s_size;\n" % (d1, b.replace(" ", "_")))
            num_modules += 1

            ##### generate the core pack/unpack kernels (more than one level)
            for d2 in gencomm.derived_types:
                OUTFILE.write("extern const unsigned char yaksuri_zei_pup_%s_%s_%s_str[];\n" % (d1, d2, b.replace(" ", "_")))
                OUTFILE.write("extern const size_t yaksuri_zei_pup_%s_%s_%s_size;\n" % (d1, d2, b.replace(" ", "_")))
                num_modules += 1
    OUTFILE.write("\n")
    OUTFILE.write("ze_module_handle_t *yaksuri_ze_modules[%d];\n\n" % num_modules)

    num_kernels = 0
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            num_kernels += 2

            ##### generate the core pack/unpack kernels (more than one level)
            for d2 in gencomm.derived_types:
                for darray in darraylist:
                    num_kernels += 2
    OUTFILE.write("ze_kernel_handle_t *yaksuri_ze_kernels[%d];\n\n" % num_kernels)

    OUTFILE.write("const char * yaksuri_zei_pup_str[%d];\n" % num_modules)
    OUTFILE.write("unsigned long yaksuri_zei_pup_size[%d];\n\n" % num_modules)

    OUTFILE.write("const char * yaksuri_zei_kernel_funcs[%d];\n" % num_kernels)
    OUTFILE.write("int yaksuri_zei_kernel_module_map[%d];\n\n" % num_kernels)

    # create modules using level zero API
    yutils.display(OUTFILE, "ze_result_t yaksuri_ze_init_module_kernel() {\n")
    OUTFILE.write("    ze_result_t zerr = ZE_RESULT_SUCCESS; \n")

    i = 0
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            OUTFILE.write("    yaksuri_zei_pup_str[%d] = yaksuri_zei_pup_%s_%s_str;\n" % (i, d1, b.replace(" ", "_")))
            OUTFILE.write("    yaksuri_zei_pup_size[%d] = yaksuri_zei_pup_%s_%s_size;\n" % (i, d1, b.replace(" ", "_")))
            i += 1
            ##### generate the core pack/unpack kernels (more than one level)
            for d2 in gencomm.derived_types:
                OUTFILE.write("    yaksuri_zei_pup_str[%d] = yaksuri_zei_pup_%s_%s_%s_str;\n" % (i, d1, d2, b.replace(" ", "_")))
                OUTFILE.write("    yaksuri_zei_pup_size[%d] = yaksuri_zei_pup_%s_%s_%s_size;\n" % (i, d1, d2, b.replace(" ", "_")))
                i += 1
    OUTFILE.write("\n")

    i = 0
    m = 0
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            for func in "pack","unpack":
                func_name = "yaksuri_zei_kernel_%s_%s_%s" % (func, d1, b.replace(" ", "_"))
                OUTFILE.write("    yaksuri_zei_kernel_funcs[%d] = \"%s\";\n" % (i, func_name))
                OUTFILE.write("    yaksuri_zei_kernel_module_map[%d] = %d;\n" % (i, m))
                i += 1
            m += 1
            ##### generate the core pack/unpack kernels (more than one level)
            for d2 in gencomm.derived_types:
                for darray in darraylist:
                    darray.append(d1)
                    darray.append(d2)
                    for func in "pack","unpack":
                        func_name = "yaksuri_zei_kernel_%s_" % func
                        for d in darray:
                            func_name = func_name + "%s_" % d
                        func_name = func_name + b.replace(" ", "_")
                        OUTFILE.write("    yaksuri_zei_kernel_funcs[%d] = \"%s\";\n" % (i, func_name))
                        OUTFILE.write("    yaksuri_zei_kernel_module_map[%d] = %d;\n" % (i, m))
                        i += 1
                    darray.pop()
                    darray.pop()
                m += 1
    OUTFILE.write("\n")

    OUTFILE.write("fn_exit:\n")
    OUTFILE.write("    return zerr; \n")
    OUTFILE.write("fn_fail:\n")
    OUTFILE.write("    goto fn_exit; \n")
    yutils.display(OUTFILE, "}\n\n")
    OUTFILE.close()

    filename = "src/backend/ze/hooks/yaksuri_zei_finalize_kernels.c"
    yutils.copyright_c(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "#include \"stdlib.h\"\n")
    yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
    yutils.display(OUTFILE, "#include \"yaksuri_zei.h\"\n")
    yutils.display(OUTFILE, "#include \"level_zero/ze_api.h\"\n\n")

    yutils.display(OUTFILE, "ze_result_t yaksuri_ze_finalize_module_kernel() {\n")
    OUTFILE.write("    ze_result_t zerr = ZE_RESULT_SUCCESS;\n")
    OUTFILE.write("    int i, k; \n\n")
    OUTFILE.write("    for (k=0; k<%d; k++) {\n" % num_kernels)
    OUTFILE.write("        if (yaksuri_ze_kernels[k] == NULL) continue;\n")
    OUTFILE.write("        for (i=0; i<yaksuri_zei_global.ndevices; i++) {\n")
    OUTFILE.write("            if (yaksuri_ze_kernels[k][i] == NULL) continue;\n")
    OUTFILE.write("            zerr = zeKernelDestroy(yaksuri_ze_kernels[k][i]);\n")
    OUTFILE.write("            if (zerr != ZE_RESULT_SUCCESS) { \n")
    OUTFILE.write("                fprintf(stderr, \"ZE Error (%s:%s,%d): %d \\n\", __func__, __FILE__, __LINE__, zerr); \n")
    OUTFILE.write("                goto fn_fail; \n");
    OUTFILE.write("            }\n");
    OUTFILE.write("            yaksuri_ze_kernels[k][i] = NULL;\n");
    OUTFILE.write("        }\n")
    OUTFILE.write("    }\n\n")
    OUTFILE.write("    for (k=0; k<%d; k++) {\n" % num_modules)
    OUTFILE.write("        if (yaksuri_ze_modules[k] == NULL) continue;\n")
    OUTFILE.write("        for (i=0; i<yaksuri_zei_global.ndevices; i++) {\n")
    OUTFILE.write("            if (yaksuri_ze_modules[k][i] == NULL) continue;\n")
    OUTFILE.write("            zerr = zeModuleDestroy(yaksuri_ze_modules[k][i]);\n")
    OUTFILE.write("            if (zerr != ZE_RESULT_SUCCESS) { \n")
    OUTFILE.write("                fprintf(stderr, \"ZE Error (%s:%s,%d): %d \\n\", __func__, __FILE__, __LINE__, zerr); \n")
    OUTFILE.write("                goto fn_fail; \n");
    OUTFILE.write("            }\n");
    OUTFILE.write("            yaksuri_ze_modules[k][i] = NULL;\n");
    OUTFILE.write("        }\n")
    OUTFILE.write("    }\n\n")
    OUTFILE.write("fn_exit:\n")
    OUTFILE.write("    return zerr; \n")
    OUTFILE.write("fn_fail:\n")
    OUTFILE.write("    goto fn_exit; \n")
    yutils.display(OUTFILE, "}\n")
    OUTFILE.close()

    ##### generate the core pack/unpack kernel declarations
    filename = "src/backend/ze/pup/yaksuri_zei_pup.h"
    yutils.copyright_c(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "#ifndef YAKSURI_ZEI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "#define YAKSURI_ZEI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "\n")

    # bit of an unnecessary header file, but creating it so we can
    # reuse the gencomm module
    k = 0
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            emptylist = [ ]
            emptylist.append(d1)
            generate_define_kernels(b, emptylist)
            emptylist.pop()
            for d2 in gencomm.derived_types:
                for darray in darraylist:
                    darray.append(d1)
                    darray.append(d2)
                    generate_define_kernels(b, darray)
                    darray.pop()
                    darray.pop()
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "#endif  /* YAKSURI_ZEI_PUP_H_INCLUDED */\n")
    OUTFILE.close()

    ##### generate the pup makefile
    filename = "src/backend/ze/pup/Makefile.pup.mk"
    yutils.copyright_makefile(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "libyaksa_la_SOURCES += \\\n")
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            yutils.display(OUTFILE, "\tsrc/backend/ze/pup/yaksuri_zei_pup_%s_%s.c \\\n" % (d1, b.replace(" ","_")))
            for d2 in gencomm.derived_types:
                yutils.display(OUTFILE, "\tsrc/backend/ze/pup/yaksuri_zei_pup_%s_%s_%s.c \\\n" % (d1, d2, b.replace(" ","_")))
    yutils.display(OUTFILE, "\tsrc/backend/ze/pup/yaksuri_zei_pup.c\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "noinst_HEADERS += \\\n")
    yutils.display(OUTFILE, "\tsrc/backend/ze/pup/yaksuri_zei_pup.h\n")
    OUTFILE.close()

    ##### generate the switching logic to select pup functions
    gencomm.populate_pupfns(args.pup_max_nesting, "ze", blklens, builtin_types, builtin_maps)




#!/usr/bin/env python

import sys
import pprint
import operator
from optparse import OptionParser
from functools import reduce

try:
	import yaml
except ImportError:
	print ("PyYAML library missing, try: yum install PyYAML")
	sys.exit(1)

# diff two list-of dicts for perf numbers
# this means just produce xfer_size, MB/sec, usec/xfer
def _diff(a,b):
	r = []

	for v1, v2 in zip(a,b):
		d = {}
		for k in list(v2.keys()):
			if k == 'xfer_size':
				d[k] = v2[k]
			elif k == 'MB/sec' or k == 'usec/xfer':
				d[k] = float((v2[k] - v1[k])) / float(v1[k]) * 100
			else:
				continue

		r.append(d.copy())

	return r

def difference(ystream):
	"""Subtract a from b and print the results"""

	# reverse stream and split based on key
	rev = reversed(list(ystream))
	a, b = {}, {}
	for i in rev:
		if not set(i).issubset(set(b)):
			b.update(i)
		else:
			a.update(i)

	result = {}
	for k in list(b.keys()):
		result[k] = _diff(a[k], b[k])

	return result

def pretty(stream):
	"""Prety-print given yaml stream and exit"""
	for i in stream: pprint.pprint(i)
	return 0

def perfprint(d):
	for k,v in list(d.items()):
		print(k, ":")
		for i in v:
			print('xfer_size: ', i['xfer_size'])
			print(', MB/sec: %.2f' % i['MB/sec'] + '%')
			print(', usec/xfer: %.2f' % i['usec/xfer'] + '%')
	

def main(argv=None):

	parser = OptionParser(description='fabtests yaml parsing utility')
	parser.add_option('-d', action='store_true', default=False, help=difference.__doc__)
	parser.add_option('-v', action='store_true', default=False, help=pretty.__doc__)
	(options, args) = parser.parse_args()

	if len(args) == 0:
		fd = sys.stdin
	elif len(args) > 1:
		class fd:
			@staticmethod
			def read():
				r1 = list(map(open, args))
				r2 = [x.read() for x in r1]
				return reduce(operator.add, r2)
	else:
		fd = open(args[0], 'r')

	yi = yaml.load_all(fd.read())

	if options.d:
		perfprint(difference(yi))

	if options.v:
		pretty(yi)

	return 0

if __name__ == "__main__":
	sys.exit(main(sys.argv))



#! /usr/bin/env python3
##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import sys
import argparse

sys.path.append('maint/')
import yutils

sys.path.append('src/backend/')
import gencomm


########################################################################################
##### Global variables
########################################################################################

num_paren_open = 0
builtin_types = [ "_Bool", "char", "wchar_t", "int8_t", "int16_t", \
                  "int32_t", "int64_t", "float", "double", "long double" ]
blklens = [ "1", "2", "3", "4", "5", "6", "7", "8", "generic" ]
builtin_maps = { }


########################################################################################
##### Type-specific functions
########################################################################################

## hvector routines
def hvector_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "int count%d = %s->u.hvector.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "int blocklength%d ATTRIBUTE((unused)) = %s->u.hvector.blocklength;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.hvector.stride;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def hvector(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (int j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    if (blklen == "generic"):
        yutils.display(OUTFILE, "for (int k%d = 0; k%d < blocklength%d; k%d++) {\n" % (suffix, suffix, suffix, suffix))
    else:
        yutils.display(OUTFILE, "for (int k%d = 0; k%d < %s; k%d++) {\n" % (suffix, suffix, blklen, suffix))
    global s
    if (last != 1):
        s += " + j%d * stride%d + k%d * extent%d" % (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + j%d * stride%d + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## blkhindx routines
def blkhindx_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "int count%d = %s->u.blkhindx.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "int blocklength%d ATTRIBUTE((unused)) = %s->u.blkhindx.blocklength;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t *restrict array_of_displs%d = %s->u.blkhindx.array_of_displs;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def blkhindx(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (int j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    if (blklen == "generic"):
        yutils.display(OUTFILE, "for (int k%d = 0; k%d < blocklength%d; k%d++) {\n" % (suffix, suffix, suffix, suffix))
    else:
        yutils.display(OUTFILE, "for (int k%d = 0; k%d < %s; k%d++) {\n" % (suffix, suffix, blklen, suffix))
    global s
    if (last != 1):
        s += " + array_of_displs%d[j%d] + k%d * extent%d" % \
             (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + array_of_displs%d[j%d] + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## hindexed routines
def hindexed_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "int count%d = %s->u.hindexed.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "int *restrict array_of_blocklengths%d = %s->u.hindexed.array_of_blocklengths;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t *restrict array_of_displs%d = %s->u.hindexed.array_of_displs;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def hindexed(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (int j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    yutils.display(OUTFILE, "for (int k%d = 0; k%d < array_of_blocklengths%d[j%d]; k%d++) {\n" % \
            (suffix, suffix, suffix, suffix, suffix))
    global s
    if (last != 1):
        s += " + array_of_displs%d[j%d] + k%d * extent%d" % \
             (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + array_of_displs%d[j%d] + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## contig routines
def contig_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "int count%d = %s->u.contig.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.contig.child->extent;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def contig(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 1
    yutils.display(OUTFILE, "for (int j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    global s
    s += " + j%d * stride%d" % (suffix, suffix)

# resized routines
def resized_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def resized(suffix, b, blklen, last):
    pass


########################################################################################
##### Core kernels
########################################################################################
def generate_kernels(b, darray, blklen):
    global num_paren_open
    global s

    # we don't need pup kernels for basic types
    if (len(darray) == 0):
        return

    # individual blocklength optimization is only for
    # hvector and blkhindx
    if (darray[-1] != "hvector" and darray[-1] != "blkhindx" and blklen != "generic"):
        return

    for func in "pack","unpack":
        ##### figure out the function name to use
        s = "int yaksuri_seqi_%s_" % func
        for d in darray:
            s = s + "%s_" % d
        # hvector and hindexed get blklen-specific function names
        if (darray[-1] != "hvector" and darray[-1] != "blkhindx"):
            s = s + b.replace(" ", "_")
        else:
            s = s + "blklen_%s_" % blklen + b.replace(" ", "_")
        yutils.display(OUTFILE, "%s(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type)\n" % s),
        yutils.display(OUTFILE, "{\n")


        ##### variable declarations
        # generic variables
        yutils.display(OUTFILE, "int rc = YAKSA_SUCCESS;\n");
        yutils.display(OUTFILE, "const char *restrict sbuf = (const char *) inbuf;\n");
        yutils.display(OUTFILE, "char *restrict dbuf = (char *) outbuf;\n");
        yutils.display(OUTFILE, "uintptr_t extent ATTRIBUTE((unused)) = type->extent;\n")
        yutils.display(OUTFILE, "\n");

        # variables specific to each nesting level
        s = "type"
        for x in range(len(darray)):
            getattr(sys.modules[__name__], "%s_decl" % darray[x])(x + 1, s, b)
            yutils.display(OUTFILE, "\n")
            s = s + "->u.%s.child" % darray[x]


        ##### non-hvector and non-blkhindx
        yutils.display(OUTFILE, "uintptr_t idx = 0;\n")
        yutils.display(OUTFILE, "for (int i = 0; i < count; i++) {\n")
        num_paren_open += 1
        s = "i * extent"
        for x in range(len(darray)):
            if (x != len(darray) - 1):
                getattr(sys.modules[__name__], darray[x])(x + 1, b, "generic", 0)
            else:
                getattr(sys.modules[__name__], darray[x])(x + 1, b, blklen, 1)

        if (func == "pack"):
            yutils.display(OUTFILE, "*((%s *) (void *) (dbuf + idx)) = *((const %s *) (const void *) (sbuf + %s));\n"
                           % (b, b, s))
        else:
            yutils.display(OUTFILE, "*((%s *) (void *) (dbuf + %s)) = *((const %s *) (const void *) (sbuf + idx));\n"
                           % (b, s, b))
        yutils.display(OUTFILE, "idx += sizeof(%s);\n" % b)
        for x in range(num_paren_open):
            yutils.display(OUTFILE, "}\n")
        num_paren_open = 0
        yutils.display(OUTFILE, "\n");
        yutils.display(OUTFILE, "return rc;\n")
        yutils.display(OUTFILE, "}\n\n")


########################################################################################
##### main function
########################################################################################
if __name__ == '__main__':
    ##### parse user arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--pup-max-nesting', type=int, default=3, help='maximum nesting levels to generate')
    args = parser.parse_args()
    if (args.pup_max_nesting < 0):
        parser.print_help()
        print
        print("===> ERROR: pup-max-nesting must be positive")
        sys.exit(1)

    ##### generate the core pack/unpack kernels (single level)
    for b in builtin_types:
        for d in gencomm.derived_types:
            filename = "src/backend/seq/pup/yaksuri_seqi_pup_%s_%s.c" % (d, b.replace(" ","_"))
            yutils.copyright_c(filename)
            OUTFILE = open(filename, "a")
            yutils.display(OUTFILE, "#include <string.h>\n")
            yutils.display(OUTFILE, "#include <stdint.h>\n")
            yutils.display(OUTFILE, "#include <wchar.h>\n")
            yutils.display(OUTFILE, "#include \"yaksuri_seqi_pup.h\"\n")
            yutils.display(OUTFILE, "\n")

            emptylist = [ ]
            emptylist.append(d)
            for blklen in blklens:
                generate_kernels(b, emptylist, blklen)
            emptylist.pop()

            OUTFILE.close()

    ##### generate the core pack/unpack kernels (more than one level)
    darraylist = [ ]
    yutils.generate_darrays(gencomm.derived_types, darraylist, args.pup_max_nesting - 2)
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            for d2 in gencomm.derived_types:
                filename = "src/backend/seq/pup/yaksuri_seqi_pup_%s_%s_%s.c" % (d1, d2, b.replace(" ","_"))
                yutils.copyright_c(filename)
                OUTFILE = open(filename, "a")
                yutils.display(OUTFILE, "#include <string.h>\n")
                yutils.display(OUTFILE, "#include <stdint.h>\n")
                yutils.display(OUTFILE, "#include <wchar.h>\n")
                yutils.display(OUTFILE, "#include \"yaksuri_seqi_pup.h\"\n")
                yutils.display(OUTFILE, "\n")

                for darray in darraylist:
                    darray.append(d1)
                    darray.append(d2)
                    for blklen in blklens:
                        generate_kernels(b, darray, blklen)
                    darray.pop()
                    darray.pop()

                OUTFILE.close()

    ##### generate the core pack/unpack kernel declarations
    filename = "src/backend/seq/pup/yaksuri_seqi_pup.h"
    yutils.copyright_c(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "#ifndef YAKSURI_SEQI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "#include <string.h>\n")
    yutils.display(OUTFILE, "#include <stdint.h>\n")
    yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
    yutils.display(OUTFILE, "\n")

    darraylist = [ ]
    yutils.generate_darrays(gencomm.derived_types, darraylist, args.pup_max_nesting)
    for b in builtin_types:
        for darray in darraylist:
            for blklen in blklens:
                # we don't need pup kernels for basic types
                if (len(darray) == 0):
                    continue

                # individual blocklength optimization is only for
                # hvector and blkhindx
                if (darray[-1] != "hvector" and darray[-1] != "blkhindx" \
                    and blklen != "generic"):
                    continue

                for func in "pack","unpack":
                    ##### figure out the function name to use
                    s = "int yaksuri_seqi_%s_" % func
                    for d in darray:
                        s = s + "%s_" % d
                    # hvector and hindexed get blklen-specific function names
                    if (len(darray) and darray[-1] != "hvector" and darray[-1] != "blkhindx"):
                        s = s + b.replace(" ", "_")
                    else:
                        s = s + "blklen_%s_" % blklen + b.replace(" ", "_")
                    yutils.display(OUTFILE, "%s" % s),
                    yutils.display(OUTFILE, "(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type);\n")

    yutils.display(OUTFILE, "#endif  /* YAKSURI_SEQI_PUP_H_INCLUDED */\n")
    OUTFILE.close()

    ##### generate the pup makefile
    filename = "src/backend/seq/pup/Makefile.pup.mk"
    yutils.copyright_makefile(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "libyaksa_la_SOURCES += \\\n")
    for b in builtin_types:
        for d1 in gencomm.derived_types:
            yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup_%s_%s.c \\\n" % \
                           (d1, b.replace(" ","_")))
            for d2 in gencomm.derived_types:
                yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup_%s_%s_%s.c \\\n" % \
                               (d1, d2, b.replace(" ","_")))
    yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seq_pup.c\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "noinst_HEADERS += \\\n")
    yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup.h\n")
    OUTFILE.close()

    ##### generate the switching logic to select pup functions
    gencomm.populate_pupfns(args.pup_max_nesting, "seq", blklens, builtin_types, builtin_maps)
